#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <stdlib.h>

typedef int8_t		s08;
typedef int16_t	s16;
typedef int32_t	s32;
typedef uint8_t	u08;
typedef uint16_t	u16;
typedef uint32_t	u32;

#define XTAL				8000000

#define DMX_StartCode		0		// DMX512 start code to react to (0 is for dimmers)
#define DMX_UART_Speed		250000	// standard DMX512 mode
#define DMX_UART_Div		(u16) (XTAL / 16 / DMX_UART_Speed - 1)
#define DMX_NUM_CHANNELS	3
#define DMX_INDICATOR_TIME	3

#define DIMMER_TimerStart	0xD8EF
#define DIMMER_AngleWidth	38
#define DIMMER_PORT			PORTA
#define DIMMER_PORT_DIR		DDRA

#define LED_PORT			PORTE
#define LED_PORT_DIR		DDRE
#define LED_Green			PE1
#define LED_Red				PE2

#define BLINK_Counter		25

#define PIN_Ext				PA0
#define PIN_Pump			PA1
#define PIN_Heater			PA2

#define PIN_Temp			ACO
#define PIN_ManFog			PA3

volatile u08 g_DMX_Status;		 				// current DMX status (global to be resettable)
volatile u08 g_DMX_Values[DMX_NUM_CHANNELS];
volatile u08 g_DMX_BaseAddress = 1;
volatile u08 g_DMX_IndicatorCounter = DMX_INDICATOR_TIME - 1;

volatile u08 g_DIMMER_PhaseCount;

volatile u08 g_DIMMER_PumpVal = 0;
volatile u08 g_DIMMER_PumpVal2 = 0;
volatile u08 g_DIMMER_ExtVal = 0;

volatile u08 g_TempState = 1;

volatile u16 g_CycleTime = 0;
volatile u08 g_OnTime = 0;

void DMX_ReadBaseAddress(void) {
 u08 temp;

 temp = PINC;
 g_DMX_BaseAddress=255-temp;
}

void DMX_init(void) {
 u08 i;
 // - UART mode
 UCSRC = (1<<URSEL) | (1<<UCSZ1) | (1<<UCSZ0); // 8n1
 UCSRB = (1<<RXCIE) | (1<<RXEN) | (1<<TXEN); // receive int and receiver enabled
 UBRRH = (u08) (DMX_UART_Div >> 8); // set baud rate
 UBRRL = (u08) (DMX_UART_Div);
 for (i=0; i<DMX_NUM_CHANNELS; i++) {
  g_DMX_Values[i] = 0;
 }
 g_DMX_Status = 0; // reset status to let ISR wait for next BREAK before doing anything
 DMX_ReadBaseAddress();
}

void DIMMER_init(void) {
 TCCR1B = 0b00000010; 					// set timer1 to ck/8
 // set interrupt 1 for phase zero-crossing control
 MCUCR |= (1 << ISC11);				// interrupt by falling edge
 MCUCR &= ~(1 << ISC10);				// <-^
 GICR |= (1 << INT1);					// enable interrupt0
 GIFR |= (1 << INTF1);					// clear interrupt0 flag
}

void init(void) {
 //ports
 DDRC = 192;
 PORTC = 255;
 DDRD = 4;
 PORTD = 8;
 LED_PORT_DIR = 255;
 LED_PORT = 255;

 DIMMER_PORT_DIR = 0b11110111;
 DIMMER_PORT =     0b00001111;

 ACSR = 0b01000000;								// bandgap reference (1.23V)

 TCCR0 = 0b00000101;							// set clk/1024
 TIMSK |= (1 << TOIE0);						// enable timer0 overflow interrupt
}

int main(void) {
 init();
 DMX_init();
 DIMMER_init();
 sei();										// Interrupts erlauben
 for (;;) {								// Endlosschleife
  g_DIMMER_ExtVal = g_DMX_Values[2];
  if (g_DMX_Values[0] > 7) {
   DIMMER_PORT &= ~(1 << PIN_Heater);			// LED an (Bit aus)
   if (g_TempState == 0) {
//    DIMMER_PORT &= ~(1 << PIN_Heater);			// LED an (Bit aus)
    g_OnTime = 0;
    g_DIMMER_PumpVal = 0;
   } else {
//    DIMMER_PORT |= 1 << PIN_Heater;				// LED aus (Bit an)
	if (bit_is_clear (PINA, PIN_ManFog)) {
     g_OnTime = 0;
     g_DIMMER_PumpVal = 255;	
	} else {
	 if (g_DMX_Values[0] < 21) {
	  g_OnTime = 0;
      g_DIMMER_PumpVal = 0;
	 } else if (g_DMX_Values[0] < 41) {
      g_CycleTime = 310;
	  g_OnTime = 10;
	 } else if (g_DMX_Values[0] < 61) {
      g_CycleTime = 370;
	  g_OnTime = 20;
	 } else if (g_DMX_Values[0] < 81) {
      g_CycleTime = 230;
	  g_OnTime = 30;
	 } else if (g_DMX_Values[0] < 101) {
      g_CycleTime = 190;
	  g_OnTime = 40;
	 } else if (g_DMX_Values[0] < 121) {
      g_CycleTime = 150;
	  g_OnTime = 50;
	 } else if (g_DMX_Values[0] < 141) {
      g_CycleTime = 135;
	  g_OnTime = 60;
	 } else if (g_DMX_Values[0] < 161) {
      g_CycleTime = 120;
	  g_OnTime = 70;
	 } else if (g_DMX_Values[0] < 181) {
      g_CycleTime = 120;
	  g_OnTime = 80;
	 } else if (g_DMX_Values[0] < 201) {
      g_CycleTime = 120;
	  g_OnTime = 90;
	 } else if (g_DMX_Values[0] < 221) {
      g_CycleTime = 120;
	  g_OnTime = 100;
	 } else if (g_DMX_Values[0] < 141) {
      g_CycleTime = 120;
	  g_OnTime = 110;
	 } else {
	  g_OnTime = 0;
      g_DIMMER_PumpVal = g_DMX_Values[1];
     }
	} 
   }	
  } else { 
   DIMMER_PORT |= 1 << PIN_Heater;				// LED aus (Bit an)
   g_DIMMER_PumpVal = 0;
  } 
 }
}

// LED-BLINKEN
SIGNAL (SIG_OVERFLOW0) {
 static u08 WaitTime = BLINK_Counter;
 static u08 SecTimer = 100;
 static u16 SecTimerPos = 0;
 static u08 TempStateCycles = 0;
 static u08 LastTempState = 0;

 if (g_OnTime > 0) {
  DIMMER_PORT |= 1 << 4;
  SecTimer--;
  if (SecTimer == 0) {
   SecTimerPos++;
   SecTimer = 100;
   if (bit_is_set (DIMMER_PORT, 7)) {
    DIMMER_PORT &= ~(1 << 7);
   } else { 
    DIMMER_PORT |= 1 << 7;
   } 
  }
  if (SecTimerPos < g_OnTime) {
   if (g_TempState == 0) {
    g_DIMMER_PumpVal = 0;
   } else {
    g_DIMMER_PumpVal = g_DMX_Values[1];
   }  
  } else {
   g_DIMMER_PumpVal = 0;
  }
  if (SecTimerPos >= g_CycleTime) {
   SecTimerPos = 0;
  }
 } else {
  DIMMER_PORT &= ~(1 << 4);
 }

 WaitTime--;
 if (WaitTime == 0) {
  // Basisadresse lesen
  DMX_ReadBaseAddress();
  if (bit_is_set (ACSR, PIN_Temp)) {
   // Heizen
   if (LastTempState == 1) {
    TempStateCycles++;
	if (TempStateCycles == 5) {
	 g_TempState = 0;
	}
   } else {
    LastTempState = 1;
    TempStateCycles = 0;
   }
  } else {
   // Temperatur erreicht
   if (LastTempState == 0) {
    TempStateCycles++;
	if (TempStateCycles == 5) {
	 g_TempState = 1;
	}
   } else {
    LastTempState = 0;
    TempStateCycles = 0;
   }
  } 
   
  // Grne LED fr Heizen / OK
  if (g_DMX_Values[0] > 7) {
   if (g_TempState == 0) {
    // Grne LED blinken lassen = heizen
    if (bit_is_set (LED_PORT, LED_Green)) {
     LED_PORT &= ~(1 << LED_Green);			// LED an (Bit aus)
    } else { 
     LED_PORT |= 1 << LED_Green;				// LED aus (Bit an)
    } 
   } else {
    // Grne LED an = Temperatur erreicht
    LED_PORT &= ~(1 << LED_Green);				// LED an (Bit aus)
   }   
  } else {
   // Grne LED aus = Gert aus
   LED_PORT |= 1 << LED_Green;				// LED aus (Bit an)
  } 
  // Rote LED fr Fehler
  if (g_DMX_IndicatorCounter > 0) {
   g_DMX_IndicatorCounter--;
   LED_PORT &= ~(1 << LED_Red);				// LED an (Bit aus)
  } else {
   g_DMX_Values[0] = 255;
   g_DMX_Values[1] = 0;
   // Rote LED blinken lassen
   if (bit_is_set (LED_PORT, LED_Red)) {
    LED_PORT &= ~(1 << LED_Red);			// LED an (Bit aus)
   } else { 
    LED_PORT |= 1 << LED_Red;				// LED aus (Bit an)
   } 
  }
  WaitTime = BLINK_Counter; 
 } 
 TCNT0 = 0xb2; // 10ms
}

// DIMMER ZEROCROSSING
SIGNAL (SIG_INTERRUPT1) {
 static u08 PumpTimerPos = 255;

 TCNT1 = DIMMER_TimerStart;				// clear angle-timer
 TIFR |= (1 << OCF1A) + (1 << TOV1);		// clear compare flag; clear overflow-flag
 TIMSK |= (1 << OCIE1A) + (1 << TOIE1);	// enable timer1 compare interrupt A; enable timer1 overflow interrupt
 GICR &= ~(1 << INT1);						// disable ext-interrupt
 g_DIMMER_PhaseCount = 0xff;
 OCR1A = DIMMER_TimerStart + DIMMER_AngleWidth;
 
 PumpTimerPos=PumpTimerPos-2;
 if (PumpTimerPos < g_DIMMER_PumpVal) {
  PumpTimerPos = 255;
  DIMMER_PORT |= 1 << 5;
  g_DIMMER_PumpVal2 = 255;
 } else {
  DIMMER_PORT &= ~(1 << 5);
  g_DIMMER_PumpVal2 = 0;
 } 
}

// DIMMER ZEROCROSSING OVERFLOW
SIGNAL (SIG_OVERFLOW1) {
 TCNT1 = DIMMER_TimerStart;				// clear angle-timer
 TIMSK |= (1 << OCIE1A);					// enable timer1 compare interrupt A
 TIMSK &= ~(1 << TOIE1);					// disable timer1 overflow interrupt
 GIFR &= ~(1 << INTF1);					// clear interrupt1 flag
 g_DIMMER_PhaseCount = 0xff;
 OCR1A = DIMMER_TimerStart + DIMMER_AngleWidth;
}

// DIMMER (TIMER COMPARE)
SIGNAL (SIG_OUTPUT_COMPARE1A) {
 u16 Temp = 0;
 u16 Current;

 if (g_DIMMER_PumpVal2 > g_DIMMER_PhaseCount) {
  DIMMER_PORT &= ~(1 << PIN_Pump);			// LED an (Bit aus)
 } else {
  DIMMER_PORT |= 1 << PIN_Pump;				// LED aus (Bit an)
 }

 if (g_DIMMER_ExtVal > g_DIMMER_PhaseCount) {
  DIMMER_PORT &= ~(1 << PIN_Ext);			// LED an (Bit aus)
 } else {
  DIMMER_PORT |= 1 << PIN_Ext;				// LED aus (Bit an)
 }
 
 do {
  g_DIMMER_PhaseCount--;
  if (g_DIMMER_PhaseCount == 0) {
   break;
  }
  Temp = OCR1A;
  Temp += DIMMER_AngleWidth;
  Current = TCNT1;
 } while (Current >= Temp);	// test if next timer1-compare-value isn't already gone
 if (g_DIMMER_PhaseCount == 0) {
  TIMSK &= ~(1 << OCIE1A);					// disable timer1 compare interrupt
  DIMMER_PORT |= (1 << PIN_Ext) | (1 << PIN_Pump);				// LED aus (Bit an)
//  DIMMER_PORT = 0b00001111;						// turn Dimmer port off (safe)
  if ((TIMSK & (1 << TOIE1)) == 0) {		// check for timer-overflow-interrupt; if timer1 overflow interrupt is enabled
   GIFR |= (1 << INTF1);					// clear interrupt1 flag
   GICR |= (1 << INT1);					// enable external interrupt1
  }
 } else {
  OCR1A = Temp;
 }
}

SIGNAL (SIG_UART_RECV) { // UART received one byte - highest priority
 static u16 DMX_Frame = 0;
 u08 Status = UCSRA; // status register must be read prior to UDR (because of 2 byte fifo buffer)
 u08 Byte = UDR; // immediately catch data from i/o register to enable reception of the next byte

 if ((Byte == 0) && (Status & (1<<FE))) { // BREAK detected (Framing Error)
  g_DMX_Status = 1;
  DMX_Frame = 0;
 } else if (DMX_Frame == 0) { // SC test
  if ((Byte == DMX_StartCode) && (g_DMX_Status == 1)) { // valid SC detected
   g_DMX_Status = 2;
  }
  DMX_Frame = 1;
 } else {
  if ((g_DMX_Status == 2) && (DMX_Frame >= g_DMX_BaseAddress) && (DMX_Frame < (g_DMX_BaseAddress + DMX_NUM_CHANNELS))) { // addressed to us
   u08 Channel = DMX_Frame - g_DMX_BaseAddress;
   if (Channel < DMX_NUM_CHANNELS) {
    g_DMX_Values[Channel] = Byte;
   }
   g_DMX_IndicatorCounter = DMX_INDICATOR_TIME - 1;
  }
  DMX_Frame++;
 }
}
